@@ -213,6 +213,7 @@ Entity NewExternalEntityN(const Char *name, int namelen, const char8 *publicid,
     e->encoding_decl = CE_unknown;
     e->standalone_decl = SDD_unspecified;
     e->ddb_filename = 0;
+    e->xml_version = XV_1_0;	/* 1.0 unless otherwise specified */
 
     e->url = 0;
 
@@ -467,6 +468,7 @@ ElementDefinition DefineElementN(Dtd dtd, const Char *name, int namelen,
     e->id_attribute = 0;
     e->xml_space_attribute = 0;
     e->xml_lang_attribute = 0;
+    e->xml_id_attribute = 0;
     e->notation_attribute = 0;
     e->cached_nsdef = 0;
     e->is_externally_declared = 0;
@@ -632,7 +634,8 @@ AttributeDefinition
     AttributeDefinition a;
     static Char xml_space[] = {'x','m','l',':','s','p','a','c','e',0}; 
     static Char xml_lang[] = {'x','m','l',':','l','a','n','g',0};
-   static Char xmlns[] = {'x','m','l','n','s',0};
+    static Char xml_id[] = {'x','m','l',':','i','d',0};
+    static Char xmlns[] = {'x','m','l','n','s',0};
     Char *t;
 
     if(!(a= Malloc(sizeof(*a))))
@@ -690,6 +693,8 @@ AttributeDefinition
 	element->xml_space_attribute = a;
     else if(Strcmp(name, xml_lang) == 0)
 	element->xml_lang_attribute = a;
+    else if(Strcmp(name, xml_id) == 0)
+	element->xml_id_attribute = a;
 
     a->cached_nsdef = 0;
 
