@@ -1,10 +1,23 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdarg.h>
+
 #if defined(WIN32) && ! defined(__CYGWIN__)
 #include <io.h>
 #include <fcntl.h>
 #endif
+
+#ifndef WIN32
+#define TIME_LIMIT
+#endif
+
+#ifdef TIME_LIMIT
+#include <signal.h>
+#include <sys/types.h>
+#include <sys/time.h>
+#include <sys/resource.h>
+#endif
+
 #include "system.h"
 #include "charset.h"
 #include "string16.h"
@@ -16,6 +29,7 @@
 #include "version.h"
 #include "namespaces.h"
 #include "infoset-print.h"
+#include "catalog.h"
 
 int attr_compare(const void *a, const void *b);
 void print_tree(Parser p, XBit bit);
@@ -24,6 +38,8 @@ void print_ns_attrs(NamespaceBinding ns, int count);
 void print_namespaces(NamespaceBinding ns);
 void print_attrs(ElementDefinition e, Attribute a);
 void print_text(Char *text);
+int printable(int c);
+void print_special(int c);
 void print_text_bit(Char *text);
 void dtd_cb(XBit bit, void *arg);
 void dtd_cb2(XBit bit, void *arg);
@@ -34,12 +50,18 @@ static const char8 *minimal_uri(const char8 *uri, const char8 *base);
 int verbose = 0, expand = 1, nsgml = 0,
     attr_defaults = 0, merge = 0, strict_xml = 0, tree = 0, validate = 0,
     xml_space = 0, namespaces = 0, simple_error = 0, experiment = 0,
-    read_dtd = 0;
-enum {o_unspec, o_none, o_bits, o_plain, o_can1, o_can2, o_can3, o_infoset} output_format = o_unspec;
-char *enc_name = 0, *base_uri = 0;
+    read_dtd = 0, unicode_check = 0, xml_id = 0;
+enum {o_unspec, o_none, o_bits, o_plain, o_can1, o_can2, o_can3, o_infoset, o_diff} output_format = o_unspec;
+char *enc_name = 0, *base_uri = 0, *my_dtd_name, *my_dtd_sysid = 0;
 CharacterEncoding encoding = CE_unknown;
 InputSource source = 0;
 int need_canonical_dtd = 0;
+int xml_version;
+
+#ifdef TIME_LIMIT
+void time_exceeded(int sig);
+int time_limit = 0;
+#endif
 
 #define canonical_output (output_format >= o_can1)
 
@@ -50,7 +72,7 @@ int main(int argc, char **argv)
 {
     int i;
     Parser p;
-    char *s;
+    char *s, *url;
     Entity ent = 0;
 
     /* Sigh... guess which well-known system doesn't have getopt() */
@@ -115,6 +137,9 @@ int main(int argc, char **argv)
 		    attr_defaults = 1;
 		    merge = 0;
 		    break;
+		case 'd':
+		    output_format = o_diff;
+		    break;
 		default:
 		    fprintf(stderr, "bad output format %s\n", argv[i]);
 		    return 1;
@@ -147,6 +172,19 @@ int main(int argc, char **argv)
 		strict_xml = 1;
 		attr_defaults = 1;
 		break;
+	    case 'R':
+#ifdef TIME_LIMIT
+		if(++i >= argc)
+		{
+		    fprintf(stderr, "-R requires argument\n");
+		    return 1;
+		}
+		time_limit = atoi(argv[i]);
+#else
+		fprintf("-R not supported on this system\n");
+		return 1;
+#endif
+		break;
 	    case 'S':
 		xml_space = 1;
 		break;
@@ -161,33 +199,105 @@ int main(int argc, char **argv)
 		}
 		base_uri = argv[i];
 		break;
+	    case 'U':
+		if(++i >= argc)
+		{
+		    fprintf(stderr, "-U requires argument\n");
+		    return 1;
+		}
+		switch(argv[i][0])
+		{
+		case '0':
+		    unicode_check = 0;
+		    break;
+		case '1':
+		    unicode_check = 1;
+		    break;
+		case '2':
+		    unicode_check = 2;
+		    break;
+		default:
+		    fprintf(stderr, "bad Unicode check level %s\n", argv[i]);
+		    return 1;
+		}
+		break;
 	    case 'd':
 		read_dtd = 1;
 		break;
+	    case 'D':
+		if(i+2 >= argc)
+		{
+		    fprintf(stderr, "-D requires 2 arguments\n");
+		    return 1;
+		}
+		my_dtd_name = argv[++i];
+		my_dtd_sysid = argv[++i];
+		break;
+	    case 'i':
+		xml_id = 1;
+		break;
+	    case 'I':
+		xml_id = 2;
+		break;
 	    case '.':
 		experiment = 1;
 		break;
 	    default:
 		fprintf(stderr, 
-			"usage: rxp [-abemnNsStvVx] [-o b|0|1|2|3] [-c encoding] [-u base_uri] [url]\n");
+			"usage: rxp [-abeiImnNRsStvVx] [-o b|0|1|2|3|i|d] [-U 0|1|2] [-c encoding] [-D name sysid] [-u base_uri] [url]\n");
 		return 1;
 	    }
     }
 
+#ifdef TIME_LIMIT
+    if(time_limit > 0)
+    {
+	struct rlimit r;
+	r.rlim_cur = r.rlim_max = time_limit;
+	signal(SIGXCPU, time_exceeded);
+	if(setrlimit(RLIMIT_CPU, &r) < 0)
+	{
+	    perror("setrlimit");
+	    return 1;
+	}
+    }
+#endif
+
     init_parser();
 
     if(verbose)
 	fprintf(stderr, "%s\n", rxp_version_string);
 
-    if(i < argc)
+    p = NewParser();
+
+    /* For use in CGI scripts, so user-supplied name is not parsed as
+       command argument */
+    url = getenv("RXPURL");
+
+    if(!url)
+	/* don't use catalogue in CGI scripts */
+	CatalogEnable(p);
+    
+    if(!url && i < argc)
+	url = argv[i];
+
+    if(url)
     {
-	ent = NewExternalEntity(0, 0, argv[i], 0, 0);
+	ent = NewExternalEntity(0, 0, url, 0, 0);
 	if(ent)
-	    source = EntityOpen(ent);
+	{
+	    if(p->entity_opener)
+		source = p->entity_opener(ent, p->entity_opener_arg);
+	    else
+		source = EntityOpen(ent);
+	}
     }
     else
 	source = SourceFromStream("<stdin>", stdin);
 
+    if(!source)
+	return 1;
+
     if(base_uri)
     {
 	/* Merge with default base URI so we can use a filename as base */
@@ -196,17 +306,16 @@ int main(int argc, char **argv)
 	free(base_uri);
     }
 
-    if(!source)
-	return 1;
-
-    p = NewParser();
-    ParserSetEntityOpener(p, entity_open);
-
     if(validate)
 	ParserSetFlag(p, Validate, 1);
     if(validate > 1)
 	ParserSetFlag(p, ErrorOnValidityErrors, 1);
 
+    if(xml_id)
+	ParserSetFlag(p, XMLID, 1);
+    if(xml_id > 1)
+	ParserSetFlag(p, XMLIDCheckUnique, 1);
+
     if(read_dtd)
     {
 	ParserSetFlag(p, TrustSDD, 0);
@@ -228,14 +337,14 @@ int main(int argc, char **argv)
     if(output_format == o_bits)
     {
 	ParserSetDtdCallback(p, dtd_cb);
-	ParserSetCallbackArg(p, p);
+	ParserSetDtdCallbackArg(p, p);
     }
 
     if(output_format == o_infoset)
     {
 	ParserSetFlag(p, ReturnNamespaceAttributes, 1);
 	ParserSetDtdCallback(p, dtd_cb2);
-	ParserSetCallbackArg(p, p);
+	ParserSetDtdCallbackArg(p, p);
     }
 
     ParserSetFlag(p, SimpleErrorFormat, simple_error);
@@ -274,12 +383,36 @@ int main(int argc, char **argv)
 	
     }
 
+    if(unicode_check)
+    {
+	ParserSetFlag(p, XML11CheckNF, 1);
+	if(unicode_check == 2)
+	    ParserSetFlag(p, XML11CheckExists, 1);
+    }
+
     if(ParserPush(p, source) == -1)
     {
 	ParserPerror(p, &p->xbit);
 	return 1;
     }
 
+    if(my_dtd_name)
+    {
+	Entity my_dtd = NewExternalEntity(0, 0, my_dtd_sysid, 0, source->entity);
+	p->dtd->name = strdup_char8_to_Char(my_dtd_name);
+	p->dtd->internal_part = 0;
+	p->dtd->external_part = my_dtd;
+	ParseDtd(p, my_dtd);
+	if(p->xbit.type == XBIT_error)
+	{
+	    if(my_dtd_sysid[0] == '-')
+		fprintf(stderr, "warning: DTD System ID starts with \"-\";"
+			        " did you forget the name argument?\n");
+	    print_bit(p, &p->xbit);
+	    return 1;
+	}
+    }
+
     if(enc_name)
     {
 	encoding = FindEncoding(enc_name);
@@ -310,12 +443,17 @@ int main(int argc, char **argv)
 		CharacterEncodingNameAndByteOrder[source->entity->encoding],
 		CharacterEncodingNameAndByteOrder[encoding]);
 
-    if(source->entity->ml_decl == ML_xml && output_format == o_plain)
+    xml_version = p->xml_version;
+
+    if((source->entity->ml_decl == ML_xml || encoding != CE_UTF_8) &&
+       output_format == o_plain)
     {
 	Printf("<?xml");
 
 	if(source->entity->version_decl)
 	    Printf(" version=\"%s\"", source->entity->version_decl);
+	else
+	    Printf(" version=\"1.0\"");
 
 	if(encoding == CE_unspecified_ascii_superset)
 	{
@@ -333,6 +471,10 @@ int main(int argc, char **argv)
 
 	Printf("?>\n");
     }
+    
+    if(source->entity->ml_decl == ML_xml && canonical_output &&
+       xml_version > XV_1_0)
+	Printf("<?xml version=\"%s\"?>", source->entity->version_decl);
 
     VectorInit(bits);
 
@@ -372,7 +514,8 @@ int main(int argc, char **argv)
 	{
 	    int status = p->seen_validity_error ? 2 : 0;
 
-	    if(output_format == o_plain)
+	    if(output_format == o_plain ||
+	       output_format == o_diff)
 		Printf("\n");
 
 	    /* Not necessary, but helps me check for leaks */
@@ -413,6 +556,7 @@ void print_tree(Parser p, XBit bit)
 	endbit.type = XBIT_end;
 	endbit.element_definition = bit->element_definition;
 	endbit.ns_element_definition = bit->ns_element_definition;
+	endbit.byte_offset = -1;
 	print_bit(p, &endbit);
     }
 }
@@ -449,8 +593,8 @@ void print_bit(Parser p, XBit bit)
 	    break;
 	case XBIT_start:
 	    if(namespaces && bit->ns_element_definition)
-		Printf("start: {%s}%S ",
-		       bit->ns_element_definition->namespace->uri,
+		Printf("start: {%S}%S ",
+		       bit->ns_element_definition->namespace->nsname,
 		       bit->element_definition->local);
 	    else
 		Printf("start: %S ", bit->element_definition->name);
@@ -462,8 +606,8 @@ void print_bit(Parser p, XBit bit)
 	    break;
 	case XBIT_empty:
 	    if(namespaces && bit->ns_element_definition)
-		Printf("empty: {%s}%S ",
-		       bit->ns_element_definition->namespace->uri,
+		Printf("empty: {%S}%S ",
+		       bit->ns_element_definition->namespace->nsname,
 		       bit->element_definition->local);
 	    else
 		Printf("empty: %S ", bit->element_definition->name);
@@ -475,8 +619,8 @@ void print_bit(Parser p, XBit bit)
 	    break;
 	case XBIT_end:
 	    if(namespaces && bit->ns_element_definition)
-		Printf("end: {%s}%S ",
-		       bit->ns_element_definition->namespace->uri,
+		Printf("end: {%S}%S ",
+		       bit->ns_element_definition->namespace->nsname,
 		       bit->element_definition->local);
 	    else
 		Printf("end: %S ", bit->element_definition->name);
@@ -615,8 +759,8 @@ void print_attrs(ElementDefinition e, Attribute a)
     {
 	if(output_format == o_bits && namespaces && 
 	   aa[i]->ns_definition && !aa[i]->ns_definition->element)
-	    Printf(" {%s}%S=\"", 
-		   aa[i]->ns_definition->namespace->uri,
+	    Printf(" {%S}%S=\"", 
+		   aa[i]->ns_definition->namespace->nsname,
 		   aa[i]->definition->local);
 	else
 	    Printf(" %S=\"", aa[i]->definition->name);
@@ -669,35 +813,23 @@ void print_text(Char *text)
 
     for(pc = last = text; *pc; pc++)
     {
-	if(*pc == '&' || *pc == '<' || *pc == '>' || *pc == '"' ||
-	   (canonical_output && (*pc == 9 || *pc == 10 || *pc == 13)))
+	int c = *pc, type = 0;
+	
+	if(c == '&' || c == '<' || c == '>' || c == '"' || 
+	   (c > 127 && !printable(c)))
+	    type = 1;
+	else if(c == 9 || c == 10 || c == 13)
+	    type = 2;
+	else if(c < 0x20 || (c >= 0x7f && c < 0xa0))
+	    type = 3;
+
+	if(type == 1 ||
+	   (canonical_output && output_format != o_diff && type == 2) ||
+	   (xml_version > XV_1_0 && type == 3))
 	{
 	    if(pc > last)
 		Printf("%.*S", pc - last, last);
-	    switch(*pc)
-	    {
-	    case '<':
-		Printf("&lt;");
-		break;
-	    case '>':
-		Printf("&gt;");
-		break;
-	    case '&':
-		Printf("&amp;");
-		break;
-	    case '"':
-		Printf("&quot;");
-		break;
-	    case 9:
-		Printf("&#9;");
-		break;
-	    case 10:
-		Printf("&#10;");
-		break;
-	    case 13:
-		Printf("&#13;");
-		break;
-	    }
+	    print_special(c);
 	    last = pc+1;
 	}
     }
@@ -706,6 +838,68 @@ void print_text(Char *text)
 	Printf("%.*S", pc - last, last);
 }
 
+int printable(int c)
+{
+    int tablenum;
+
+    switch(encoding)
+    {
+    case CE_unspecified_ascii_superset:
+    case CE_ISO_646:
+    default:
+	return c <= 127;
+
+    case CE_ISO_8859_1:
+	return c <= 255;
+
+    case CE_ISO_8859_2:
+    case CE_ISO_8859_3:
+    case CE_ISO_8859_4:
+    case CE_ISO_8859_5:
+    case CE_ISO_8859_6:
+    case CE_ISO_8859_7:
+    case CE_ISO_8859_8:
+    case CE_ISO_8859_9:
+    case CE_ISO_8859_10:
+    case CE_ISO_8859_11:
+    case CE_ISO_8859_13:
+    case CE_ISO_8859_14:
+    case CE_ISO_8859_15:
+	tablenum = (encoding - CE_ISO_8859_2);
+	return c <= iso_max_val[tablenum] && unicode_to_iso[tablenum][c] != '?';
+
+    case CE_UTF_8:
+    case CE_UTF_16B:
+    case CE_UTF_16L:
+	return 1;
+
+    case CE_ISO_10646_UCS_2B:
+    case CE_ISO_10646_UCS_2L :
+	return c <= 0xffff;
+    }
+}
+
+void print_special(int c)
+{
+    switch(c)
+    {
+    case '<':
+	Printf("&lt;");
+	break;
+    case '>':
+	Printf("&gt;");
+	break;
+    case '&':
+	Printf("&amp;");
+	break;
+    case '"':
+	Printf("&quot;");
+	break;
+    default:
+	Printf(canonical_output ? "&#%d;" : "&#x%02x;", c);
+    }
+}
+
 InputSource entity_open(Entity ent, void *arg)
 {
     if(ent->publicid && 
@@ -730,6 +924,7 @@ InputSource entity_open(Entity ent, void *arg)
 void print_ns_attrs(NamespaceBinding ns, int count)
 {
     NamespaceBinding n;
+    static Char empty[] = {0};
 
     if(!namespaces)
 	return;
@@ -739,15 +934,17 @@ void print_ns_attrs(NamespaceBinding ns, int count)
 	/* Don't need to worry about duplicates, because that could only
 	   happen if there were repeated attributes. */
 	if(n->prefix)
-	    Printf(" xmlns:%S=\"%s\"", n->prefix, n->namespace->uri);
+	    Printf(" xmlns:%S=\"%S\"", n->prefix,
+		   n->namespace ? n->namespace->nsname : empty);
 	else
-	    Printf(" xmlns=\"%s\"", n->namespace ? n->namespace->uri : "");
+	    Printf(" xmlns=\"%S\"", n->namespace ? n->namespace->nsname : empty);
     }
 }
 
 void print_namespaces(NamespaceBinding ns)
 {
     NamespaceBinding m, n;
+    static Char null[] = {'[','n','u','l','l',']',0};
 
     if(!namespaces)
 	return;
@@ -762,9 +959,9 @@ void print_namespaces(NamespaceBinding ns)
 		goto done;
 	}
 	if(n->prefix)
-	    Printf(" %S->%s", n->prefix, n->namespace->uri);
+	    Printf(" %S->%S", n->prefix, n->namespace ? n->namespace->nsname : null);
 	else
-	    Printf(" [default]->%s", n->namespace ? n->namespace->uri : "[null]");
+	    Printf(" [default]->%S", n->namespace ? n->namespace->nsname : null);
 
     done:
 	;
@@ -899,3 +1096,12 @@ static const char8 *minimal_uri(const char8 *uri, const char8 *base)
 
     return u+1;
 }
+
+#ifdef TIME_LIMIT
+void time_exceeded(int sig)
+{
+    fprintf(stderr, "CPU time limit (%d seconds) exceeded, sorry\n", 
+	    time_limit);
+    exit(1);
+}
+#endif
