@@ -203,6 +203,11 @@ static int ConvertASCII(const char8 *buf, int count, FILE16 *file)
     case CE_ISO_8859_7:
     case CE_ISO_8859_8:
     case CE_ISO_8859_9:
+    case CE_ISO_8859_10:
+    case CE_ISO_8859_11:
+    case CE_ISO_8859_13:
+    case CE_ISO_8859_14:
+    case CE_ISO_8859_15:
     case CE_unspecified_ascii_superset:
 	if(file->flags & FILE16_crlf)
 	{
@@ -310,6 +315,11 @@ static int ConvertUTF16(const char16 *buf, int count, FILE16 *file)
     case CE_ISO_8859_7:
     case CE_ISO_8859_8:
     case CE_ISO_8859_9:
+    case CE_ISO_8859_10:
+    case CE_ISO_8859_11:
+    case CE_ISO_8859_13:
+    case CE_ISO_8859_14:
+    case CE_ISO_8859_15:
 	tablenum = (file->enc - CE_ISO_8859_2);
 	max = iso_max_val[tablenum];
 	from_unicode = unicode_to_iso[tablenum];
@@ -629,7 +639,7 @@ int Vfprintf(FILE16 *file, const char *format, va_list args)
     char16 cbuf[1];
 #endif
     int mflag, pflag, sflag, hflag, zflag;
-    int l, h, L;
+    int l, h, L, ll;
     int nchars = 0;
 
     while((c = *format++))
@@ -644,7 +654,7 @@ int Vfprintf(FILE16 *file, const char *format, va_list args)
 	width = 0;
 	prec = -1;
 	mflag=0, pflag=0, sflag=0, hflag=0, zflag=0;
-	l=0, h=0, L=0;
+	l=0, h=0, L=0, ll=0;
 
 	while(1)
 	{
@@ -706,6 +716,14 @@ int Vfprintf(FILE16 *file, const char *format, va_list args)
 	case 'l':
 	    l = 1;
 	    c = *format++;
+#ifdef HAVE_LONG_LONG
+	    if(c == 'l')
+	    {
+		l = 0;
+		ll = 1;
+		c = *format++;
+	    }
+#endif
 	    break;
 	case 'h':
 	    h = 1;
@@ -746,6 +764,10 @@ int Vfprintf(FILE16 *file, const char *format, va_list args)
 		sprintf(val, fmt, va_arg(args, int)); /* promoted to int */
 	    else if(l)
 		sprintf(val, fmt, va_arg(args, long));
+#ifdef HAVE_LONG_LONG
+	    else if(ll)
+		sprintf(val, fmt, va_arg(args, long long));
+#endif
 	    else
 		sprintf(val, fmt, va_arg(args, int));
 	    for(p=val; *p; p++)
@@ -804,12 +826,11 @@ int Vfprintf(FILE16 *file, const char *format, va_list args)
 		    return -1;
 		count = 0;
 		nchars += n;
-		while(n > 0)
+		for(i = n; i > 0; i -= BufferSize)
 		{
 		    /* ConvertUTF16 can only handle <= BufferSize chars */
-		    if(ConvertUTF16(q, n > BufferSize ? BufferSize : n, file) == -1)
+		    if(ConvertUTF16(q, i > BufferSize ? BufferSize : i, file) == -1)
 			return -1;
-		    n -= BufferSize;
 		    q += BufferSize;
 		}
 	    }
